from pathlib import Path
from unittest.mock import Mock

import pytest
import requests


BASE_PUML_URL = "https://mocked.org/"
BASE_PUML_KEYWORD = "puml"
CUSTOM_PUML_KEYWORD = "plantuml"
TESTDATA_DIR = Path(__file__).resolve().parent.joinpath("testdata")


@pytest.fixture(scope="package")
def diagram_and_encoded():
    """The fixture to return puml diagram and
    the encoded by plantuml.com string
    """
    return (
        "@startuml\nBob -> Alice : hello\n@enduml",
        "SoWkIImgAStDuNBAJrBGjLDmpCbCJbMmKiX8pSd9vt98pKi1IW80",
    )


@pytest.fixture(scope="package")
def svg_diagram():
    with open(TESTDATA_DIR.joinpath("plantuml.svg")) as f:
        return f.read()


@pytest.fixture(scope="function")
def mock_requests(monkeypatch, svg_diagram):
    p = Mock()
    p.return_value.content = svg_diagram.encode("utf-8")
    monkeypatch.setattr(requests, "get", p)
    yield p



import uuid
from unittest.mock import MagicMock

import jinja2
import pytest

from uuid import UUID

from mkdocs.config.config_options import Config

from mkdocs_puml.plugin import PlantUMLPlugin
from tests.conftest import BASE_PUML_URL, CUSTOM_PUML_KEYWORD, TESTDATA_DIR


def is_uuid_valid(uuid_str: str) -> bool:
    """Check whether uuid_str is valid uuid4

    Args:
        uuid_str: uuid to test

    Returns:
        Boolean designating if uuid_str is valid
    """
    try:
        uuid_obj = UUID(uuid_str, version=4)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_str


@pytest.fixture
def plugin_config():
    c = Config(schema=PlantUMLPlugin.config_scheme)
    c["puml_url"] = BASE_PUML_URL
    c["auto_dark"] = False
    c["extra_css"] = []
    return c


@pytest.fixture
def plugin_config_custom_keyword(plugin_config):
    plugin_config["puml_keyword"] = CUSTOM_PUML_KEYWORD
    return plugin_config


@pytest.fixture
def plugin_environment():
    loader = jinja2.FileSystemLoader(searchpath=TESTDATA_DIR)
    return jinja2.Environment(loader=loader)


@pytest.fixture
def plant_uml_plugin(plugin_config):
    plugin = PlantUMLPlugin()
    plugin.config = plugin_config
    plugin.on_config(plugin_config)

    return plugin


@pytest.fixture
def plant_uml_plugin_custom_keyword(plugin_config_custom_keyword):
    plugin = PlantUMLPlugin()
    plugin.config = plugin_config_custom_keyword
    plugin.on_config(plugin_config_custom_keyword)

    return plugin


@pytest.fixture
def plant_uml_plugin_dark(plugin_config_custom_keyword: Config):
    plugin = PlantUMLPlugin()
    plugin_config_custom_keyword["auto_dark"] = True
    plugin.config = plugin_config_custom_keyword
    plugin.on_config(plugin_config_custom_keyword)

    return plugin


@pytest.fixture
def diagrams_dict(diagram_and_encoded):
    return {
        str(uuid.uuid4()): diagram_and_encoded[0],
        str(uuid.uuid4()): diagram_and_encoded[0],
        str(uuid.uuid4()): diagram_and_encoded[0],
    }


@pytest.fixture(scope="package")
def md_lines():
    with open(TESTDATA_DIR.joinpath("markdown.md")) as f:
        return f.readlines()


@pytest.fixture
def html_page(plugin_environment, diagrams_dict):
    page = MagicMock(title="Test, page", file=MagicMock(), config=MagicMock())
    template = plugin_environment.get_template("output.html")
    page.content = template.render(
        uuid_class=PlantUMLPlugin.pre_class_name, uuids=diagrams_dict.keys()
    )
    page.html = page.content
    return page



from mkdocs_puml.plugin import PlantUMLPlugin
from mkdocs_puml.puml import PlantUML
from tests.conftest import BASE_PUML_KEYWORD, BASE_PUML_URL, CUSTOM_PUML_KEYWORD
from tests.plugins.conftest import is_uuid_valid


def test_on_config(plugin_config):
    # Test if the plugin is correctly configured with default settings
    plugin = PlantUMLPlugin()
    plugin.config = plugin_config

    plugin.on_config(plugin_config)

    assert isinstance(plugin.puml_light, PlantUML)
    assert isinstance(plugin.puml_dark, PlantUML)
    assert plugin.puml_light.base_url == BASE_PUML_URL + "svg/"
    assert plugin.puml_dark.base_url == BASE_PUML_URL + "dsvg/"
    assert plugin.puml_keyword == BASE_PUML_KEYWORD
    assert "assets/stylesheets/puml.css" in plugin_config['extra_css']


def test_on_config_custom_keyword(plugin_config_custom_keyword):
    # Test if the plugin is correctly configured with a custom keyword
    plugin = PlantUMLPlugin()
    plugin.config = plugin_config_custom_keyword

    plugin.on_config(plugin_config_custom_keyword)

    assert isinstance(plugin.puml_light, PlantUML)
    assert isinstance(plugin.puml_dark, PlantUML)
    assert plugin.puml_light.base_url == BASE_PUML_URL + "svg/"
    assert plugin.puml_dark.base_url == BASE_PUML_URL + "dsvg/"
    assert plugin.puml_keyword == CUSTOM_PUML_KEYWORD


def test_on_page_markdown(plant_uml_plugin, md_lines):
    # Test if PlantUML diagrams are correctly extracted from markdown
    plant_uml_plugin.on_page_markdown("\n".join(md_lines))

    assert len(plant_uml_plugin.diagrams) == 2

    for key in plant_uml_plugin.diagrams.keys():
        assert is_uuid_valid(key)

    for val in plant_uml_plugin.diagrams.values():
        assert "@startuml" in val and "@enduml" in val


def test_on_page_markdown_custom_keyword(plant_uml_plugin_custom_keyword, md_lines):
    # Test if PlantUML diagrams are correctly extracted with a custom keyword
    plant_uml_plugin_custom_keyword.on_page_markdown("\n".join(md_lines))

    assert len(plant_uml_plugin_custom_keyword.diagrams) == 1

    for key in plant_uml_plugin_custom_keyword.diagrams.keys():
        assert is_uuid_valid(key)

    for val in plant_uml_plugin_custom_keyword.diagrams.values():
        assert "@startuml" in val and "@enduml" in val


def test_on_env(mock_requests, plant_uml_plugin, diagrams_dict, plugin_environment):
    # Test if PlantUML diagrams are correctly converted to SVG
    plant_uml_plugin.diagrams = diagrams_dict
    plant_uml_plugin.on_env(plugin_environment)

    assert mock_requests.call_count == len(diagrams_dict)

    for light_svg, dark_svg in plant_uml_plugin.diagrams.values():
        assert isinstance(light_svg, str)
        assert light_svg.startswith("<svg")
        assert dark_svg is None  # Since auto_dark is False by default


def test_on_env_auto_dark(
    mock_requests, plant_uml_plugin_dark, diagrams_dict, plugin_environment
):
    # Test if PlantUML diagrams are correctly converted to both light and dark SVGs
    plant_uml_plugin_dark.diagrams = diagrams_dict
    plant_uml_plugin_dark.on_env(plugin_environment)

    assert mock_requests.call_count == 2 * len(diagrams_dict)

    for light_svg, dark_svg in plant_uml_plugin_dark.diagrams.values():
        assert isinstance(light_svg, str)
        assert isinstance(dark_svg, str)
        assert light_svg.startswith("<svg")
        assert dark_svg.startswith("<svg")


def test_on_post_page(plant_uml_plugin, diagrams_dict, html_page):
    # Test if PlantUML diagrams are correctly inserted into the HTML output

    # After on_env(...) method diagrams attribute changes its
    # structure to {key: (light_value, dark_value | None)}
    plant_uml_plugin.diagrams = {k: (v, None) for k, v in diagrams_dict.items()}
    output = plant_uml_plugin.on_post_page(html_page.content, html_page)

    assert output.count('<div class="puml light">') == len(diagrams_dict)
    # assert '<script src="assets/javascripts/puml/dark.js"></script>' in output

    # Test the case where page.html exists
    html_page.html = html_page.content
    output = plant_uml_plugin.on_post_page(html_page.content, html_page)
    assert html_page.html.count('<div class="puml light">') == len(diagrams_dict)


def test_on_post_page_without_html_attribute(
    plant_uml_plugin, diagrams_dict, html_page
):
    # Test if the plugin handles pages without the 'html' attribute correctly
    plant_uml_plugin.diagrams = {k: (v, None) for k, v in diagrams_dict.items()}
    delattr(html_page, "html")
    output = plant_uml_plugin.on_post_page(html_page.content, html_page)

    assert output.count('<div class="puml light">') == len(diagrams_dict)


def test_on_post_build(tmp_path, plant_uml_plugin):
    # Test if static files are correctly copied during the build process
    config = {"site_dir": str(tmp_path)}
    dest_dir = tmp_path.joinpath("assets/stylesheets")

    plant_uml_plugin.on_post_build(config)

    assert dest_dir.joinpath("puml.css").exists()


def test_on_post_build_with_subdirectory(tmp_path, plant_uml_plugin):
    # Test if the plugin correctly handles subdirectories in the static folder
    config = {"site_dir": str(tmp_path)}

    plant_uml_plugin.on_post_build(config)

    dest_dir = tmp_path.joinpath("assets/stylesheets")
    assert dest_dir.joinpath("puml.css").exists()


def test_replace_method(plant_uml_plugin):
    # Test if the _replace method correctly handles both light and dark SVGs
    key = "test_key"
    content = f'<pre class="{plant_uml_plugin.pre_class_name}">{key}</pre>'

    # Test without dark SVG
    plant_uml_plugin.diagrams[key] = ("<svg>light</svg>", None)
    result = plant_uml_plugin._replace(key, content)
    assert '<div class="puml light"><svg>light</svg></div>' in result

    # Test with dark SVG
    plant_uml_plugin.diagrams[key] = ("<svg>light</svg>", "<svg>dark</svg>")
    result = plant_uml_plugin._replace(key, content)
    assert '<div class="puml light"><svg>light</svg></div>' in result
    assert '<div class="puml dark"><svg>dark</svg></div>' in result


def test_unsupported_output(mock_requests, plant_uml_plugin, diagrams_dict, plugin_environment):
    # All requests return an erroneous response
    mock_requests.return_value.ok = False

    # Try converting to SVG
    plant_uml_plugin.auto_dark = True
    plant_uml_plugin.diagrams = diagrams_dict
    plant_uml_plugin.on_env(plugin_environment)

    for diagram in plant_uml_plugin.diagrams.values():
        for variant in diagram:
            assert '<text>Error</text>' in variant



import logging
import re
import typing
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urljoin
from xml.dom.minidom import Element, parseString  # nosec

import requests

from mkdocs_puml.encoder import encode


logger = logging.getLogger("mkdocs.plugins.plantuml")


class PlantUML:
    """PlantUML converter class.
    It requests PUML service, updates received `svg`
    and returns it to the user.

    Attributes:
        base_url (str): Base URL to the PUML service
        num_workers (int): The size of pool to run requests in
        verify_ssl (bool): Designates whether the ``requests`` should verify SSL certiticate
        output_format (str): The output format for the diagrams (e.g., "svg" or "dsvg")

    Examples:
        Use this class as::

            puml = PlantUML("https://www.plantuml.com")
            svg = puml.translate([diagram])[0]
    """

    _html_comment_regex = re.compile(r"<!--.*?-->", flags=re.DOTALL)
    ERROR_SVG = "<svg><text>Error</text></svg>"

    def __init__(
        self,
        base_url: str,
        num_workers: int = 5,
        verify_ssl: bool = True,
        output_format: str = "svg",
    ):
        self.base_url = base_url if base_url.endswith("/") else f"{base_url}/"
        self.base_url = f"{self.base_url}{output_format}/"

        if num_workers <= 0:
            raise ValueError("`num_workers` argument should be bigger than 0.")
        self.num_workers = num_workers
        self.verify_ssl = verify_ssl

    def translate(self, diagrams: typing.Iterable[str]) -> typing.List[str]:
        """Translate string diagram into HTML div
        block containing the received SVG image.

        Examples:
                This method translates content
                into <svg> image of the diagram

        Args:
            diagrams (list): string representation of PUML diagram

        Returns:
            SVG image of built diagram
        """
        encoded = [self.preprocess(v) for v in diagrams]

        with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            svg_images = executor.map(self.request, encoded)

        return [self.postprocess(v) for v in svg_images]

    def preprocess(self, content: str) -> str:
        """Preprocess the content before pass it
        to the plantuml service.

        Encoding of the content should be
        done in the step of preprocessing.

        Args:
            content (str): string representation PUML diagram
        Returns:
            Preprocessed PUML diagram
        """
        return encode(content)

    def postprocess(self, content: str) -> str:
        """Postprocess the received from plantuml service
        SVG diagram.

        Potentially, here could be the code
        that applies CSS styling to the SVG.

        Args:
            content (str): SVG representation of build diagram
        Returns:
            Postprocessed SVG diagram
        """
        diagram_content = self._clean_comments(content)

        svg = self._convert_to_dom(diagram_content)
        self._stylize_svg(svg)

        return svg.toxml()

    def request(self, encoded_diagram: str) -> str:
        """Request plantuml service with the encoded diagram;
        return SVG content

        Args:
            encoded_diagram (str): Encoded string representation of the diagram
        Returns:
            SVG representation of the diagram
        """
        resp = requests.get(
            urljoin(self.base_url, encoded_diagram), verify=self.verify_ssl
        )

        if not resp.ok:
            logger.warning(
                f"While building diagram \n\n{encoded_diagram}\n\nServer responded"
                f" with a status {resp.status_code}"
            )
            return self.ERROR_SVG

        # Use 'ignore' to strip non-utf chars
        return resp.content.decode("utf-8", errors="ignore")

    def _clean_comments(self, content: str) -> str:
        return self._html_comment_regex.sub("", content)

    def _convert_to_dom(self, content: str) -> Element:
        """The method to convert received SVG into XML DOM
        for future modifications
        """
        dom = parseString(content)  # nosec
        svg = dom.getElementsByTagName("svg")[0]
        return svg

    def _stylize_svg(self, svg: Element):
        """This method is used for SVG tags modifications.

        Notes:
            It can be used to add support of light / dark theme.
        """
        svg.setAttribute("preserveAspectRatio", "xMidYMid meet")
        svg.setAttribute("style", "background: var(--md-default-bg-color)")



import pytest

from mkdocs_puml.puml import PlantUML
from tests.conftest import BASE_PUML_URL


def test_url_with_slash():
    # Verify base_url ends with slash when provided with trailing slash
    puml = PlantUML(BASE_PUML_URL)
    assert puml.base_url.endswith("/")


def test_url_without_slash():
    # Ensure base_url ends with slash when provided without trailing slash
    puml = PlantUML(BASE_PUML_URL[:-1])
    assert puml.base_url.endswith("/")


def test_num_worker_less_or_equal_zero():
    # Confirm ValueError raised for invalid num_workers
    with pytest.raises(ValueError, match="`num_workers` argument should be bigger than 0."):
        PlantUML(BASE_PUML_URL, num_workers=0)

    with pytest.raises(ValueError, match="`num_workers` argument should be bigger than 0."):
        PlantUML(BASE_PUML_URL, num_workers=-1)


def test_translate(diagram_and_encoded: tuple[str, str], mock_requests):
    # Verify translation of multiple diagrams to SVG
    diagram, encoded = diagram_and_encoded

    diagrams = [diagram] * 2

    puml = PlantUML(BASE_PUML_URL)
    resp = puml.translate(diagrams)

    assert puml.base_url.endswith("/")

    assert len(resp) == 2

    for r in resp:
        assert r.startswith("<svg")
        assert not puml._html_comment_regex.search(r)
        assert 'preserveAspectRatio="xMidYMid meet"' in r



from mkdocs_puml.encoder import encode


def test_encode(diagram_and_encoded: tuple[str, str]):
    # Ensures the encoded output matches the expected result.
    diagram, expected = diagram_and_encoded
    encoded = encode(diagram)

    assert encoded == expected



from pathlib import Path
import typing
import re
import uuid
import os
import shutil

from mkdocs.config.base import Config
from mkdocs.config.config_options import Type
from mkdocs.plugins import BasePlugin

from mkdocs_puml.puml import PlantUML


class PlantUMLPlugin(BasePlugin):
    """MKDocs plugin that converts puml diagrams into SVG images.

    It works only with a remote PlantUML service. You should add
    these configs into ``mkdocs.yml``::

            plugins:
                - mkdocs_puml:
                    puml_url: https://www.plantuml.com/plantuml
                    num_workers: 10

    Attributes:
        pre_class_name (str): the class that will be set to intermediate <pre> tag
                              containing uuid code
        config_scheme (str): config scheme to set by user in mkdocs.yml file

        regex (re.Pattern): regex to find all puml code blocks
        uuid_regex (re.Pattern): regex to find all uuid <pre> blocks
        puml (PlantUML): PlantUML instance that requests PlantUML service
        diagrams (dict): Dictionary containing the diagrams (puml and later svg) and their keys
        puml_keyword (str): keyword used to find PlantUML blocks within Markdown files
        verify_ssl (bool): Designates whether the ``requests`` should verify SSL certificate
        auto_dark (bool): Designates whether the plugin should automatically generate dark mode images.
    """

    pre_class_name = "diagram-uuid"

    config_scheme = (
        ("puml_url", Type(str, required=True)),
        ("num_workers", Type(int, default=8)),
        ("puml_keyword", Type(str, default="puml")),
        ("verify_ssl", Type(bool, default=True)),
        ("auto_dark", Type(bool, default=True)),
    )

    def __init__(self):
        self.regex: typing.Optional[typing.Any] = None
        self.uuid_regex = re.compile(
            rf'<pre class="{self.pre_class_name}">(.+?)</pre>', flags=re.DOTALL
        )

        self.puml: typing.Optional[PlantUML] = None
        self.diagrams = {
            # key - uuid: value - puml. After on_env — svg
        }

    def on_config(self, config: Config) -> Config:
        """Event that is fired by mkdocs when configs are created.

        self.puml_light, self.puml_dark instances are populated in this event.
        Also, `puml.css` that enable dark / light mode styles is added to `extra_css`.

        Args:
            config: Full mkdocs.yml config file. To access configs of PlantUMLPlugin only,
                    use self.config attribute.

        Returns:
            Full config of the mkdocs
        """
        config['extra_css'].append("assets/stylesheets/puml.css")
        self.puml_light = PlantUML(
            self.config["puml_url"],
            num_workers=self.config["num_workers"],
            verify_ssl=self.config["verify_ssl"],
            output_format="svg",
        )
        self.puml_dark = PlantUML(
            self.config["puml_url"],
            num_workers=self.config["num_workers"],
            verify_ssl=self.config["verify_ssl"],
            output_format="dsvg",
        )
        self.puml_keyword = self.config["puml_keyword"]
        self.regex = re.compile(rf"```{self.puml_keyword}(\n.+?)```", flags=re.DOTALL)
        self.auto_dark = self.config["auto_dark"]
        return config

    def on_page_markdown(self, markdown: str, *args, **kwargs) -> str:
        """Event to fire for each .md page.

        Here, all ``puml`` code blocks are found and added to self.diagrams
        with the corresponding uuid key.

        Then, <pre class="...">{uuid of diagram}</pre> tags are added to
        the markdown page.

        Args:
            markdown: Markdown page in which to look for ``puml`` diagrams.

        Returns:
            Updated markdown page
        """
        schemes = self.regex.findall(markdown)

        for v in schemes:
            id_ = str(uuid.uuid4())
            self.diagrams[id_] = v
            markdown = markdown.replace(
                f"```{self.puml_keyword}{v}```", f'<pre class="{self.pre_class_name}">{id_}</pre>'
            )

        return markdown

    def on_env(self, env, *args, **kwargs):
        """The event is fired when jinja environment is configured.
        Such as it is fired once when all .md pages are processed,
        we can use it to request PlantUML service and convert the
        diagrams.

        Args:
            env: jinja environment
        Returns:
            Jinja environment
        """
        # TODO: self.diagrams changes its structure throughout a program run:
        #       1. Initially it's {key: "value"}
        #       2. After translate, it's {key: ("light_svg", "dark_svg" | None)}
        #       3. Align to a single format
        diagram_contents = [diagram for diagram in self.diagrams.values()]

        if self.auto_dark:
            light_svgs = self.puml_light.translate(diagram_contents)
            dark_svgs = self.puml_dark.translate(diagram_contents)
            for key, light_svg, dark_svg in zip(self.diagrams, light_svgs, dark_svgs):
                self.diagrams[key] = (light_svg, dark_svg)
        else:
            svgs = self.puml_light.translate(diagram_contents)
            for key, svg in zip(self.diagrams, svgs):
                self.diagrams[key] = (svg, None)
        return env

    def on_post_page(self, output: str, page, *args, **kwargs) -> str:
        """The event is fired after HTML page is rendered.
        Here, we substitute <pre> tags with the corresponding SVG images.

        Args:
            output: rendered HTML in str format
            page: Page object

        Returns:
            HTML page containing SVG diagrams
        """
        schemes = self.uuid_regex.findall(output)
        for v in schemes:
            output = self._replace(v, output)
            page.content = output

            # MkDocs >=1.4 doesn't have html attribute.
            # This is required for integration with mkdocs-print-page plugin.
            # TODO: Remove the support of older versions in future releases
            if hasattr(page, "html") and page.html is not None:
                page.html = output

        return output

    def _replace(self, key: str, content: str) -> str:
        """Replace a UUID key with a real diagram in a
        content
        """
        light_svg, dark_svg = self.diagrams[key]
        if dark_svg:
            replacement = (
                f'<div class="puml light">{light_svg}</div>'
                f'<div class="puml dark">{dark_svg}</div>'
            )
        else:
            replacement = f'<div class="puml light">{light_svg}</div>'
        return content.replace(f'<pre class="{self.pre_class_name}">{key}</pre>', replacement)

    def on_post_build(self, config):
        """
        Event triggered after the build process is complete.

        This method is responsible for copying static files from the plugin's
        `static` directory to the specified `assets/javascripts/puml` directory
        in the site output. This ensures that the necessary JavaScript files
        are available in the final site.

        Args:
            config (dict): The MkDocs configuration object.

        """
        # Path to the static directory in the plugin
        puml_css = Path(__file__).parent.joinpath("static/puml.css")
        # Destination directory in the site output
        dest_dir = Path(config["site_dir"]).joinpath("assets/stylesheets/")

        if not dest_dir.exists():
            os.makedirs(dest_dir)

        shutil.copy(puml_css, dest_dir)



import base64
import string
from zlib import compress

__all__ = ("encode",)

_B64_CHARS = f"{string.ascii_uppercase}{string.ascii_lowercase}{string.digits}+/"
_PUML_CHARS = f"{string.digits}{string.ascii_uppercase}{string.ascii_lowercase}-_"

_TRANSLATE_MAP = bytes.maketrans(_B64_CHARS.encode("utf-8"), _PUML_CHARS.encode("utf-8"))


def encode(content: str) -> str:
    """Encode plantUML diagram into a string
    that can be used in a communication with
    plantUML service.

    Read more at https://plantuml.com/en/text-encoding

    Args:
        content (str): plantUML diagram in a string representation

    Returns:
        Encoded string that can be used in plantUML service
        to build diagram images
    """
    content = compress(content.encode("utf-8"))[2:-4]  # 0:2 - header, -4: - checksum
    content = base64.b64encode(content)
    content = content.translate(_TRANSLATE_MAP)
    return content.decode("utf-8")
